"""Utility functions for testing twill."""

import getpass
import os
import subprocess
import sys
import time
from io import StringIO
from pathlib import Path
from typing import Optional, TextIO

import twill

test_dir = Path(__file__).parent  # test directory
current_dir = Path.cwd()  # current working directory

HOST = "127.0.0.1"  # interface to run the server on
PORT = 8080  # default port to run the server on
SLEEP = 0.5  # time to wait for the server to start

START = True  # whether to automatically start the test server
LOG = None  # name of the server log file or None

_url = None  # current server url
__popen__ = None  # server process


def get_url() -> str:
    """Get the current server URL."""
    if _url is None:
        raise RuntimeError("Server has not yet been started!")
    return _url


def cd_test_dir() -> None:
    """Make the test directory the current directory."""
    os.chdir(test_dir)


def pop_test_dir() -> None:
    """Restore the current directory before running the tests."""
    os.chdir(current_dir)


def mock_getpass(
    prompt: str = "Password: ",  # noqa: ARG001
    stream: Optional[TextIO] = None,  # noqa: ARG001
) -> str:
    """Mock getpass function."""
    return "pass"


def execute_script(
    filename: str, inp: Optional[str] = None, initial_url: Optional[str] = None
) -> None:
    """Execute twill script with the given filename."""
    if filename != "-":
        filename = str(Path(test_dir, filename))

    if inp:
        # use inp as the stdin for the actual script commands
        stdin, sys.stdin = sys.stdin, StringIO(inp)
        real_getpass, getpass.getpass = getpass.getpass, mock_getpass
    try:
        twill.execute_file(filename, initial_url=initial_url)
    finally:
        if inp:
            sys.stdin = stdin
            getpass.getpass = real_getpass


def execute_shell(
    filename: str,
    inp: Optional[str] = None,
    initial_url: Optional[str] = None,
    *,
    fail_on_unknown: bool = False,
) -> None:
    """Execute twill script with the given filename using the shell."""
    # use filename as the stdin *for the shell object only*
    if filename != "-":
        filename = str(Path(test_dir, filename))

    with open(filename, encoding="utf-8") as cmd_file:
        cmd_content = cmd_file.read()
    cmd_content += "\nquit\n"
    cmd_inp = StringIO(cmd_content)

    if inp:
        # use inp as the std input for the actual script commands
        stdin, sys.stdin = sys.stdin, StringIO(inp)
        real_getpass, getpass.getpass = getpass.getpass, mock_getpass
    try:
        loop = twill.shell.TwillCommandLoop(
            initial_url=initial_url,
            stdin=cmd_inp,
            fail_on_unknown=fail_on_unknown,
        )
        loop.cmdloop()
    except SystemExit:
        pass
    finally:
        if inp:
            sys.stdin = stdin
            getpass.getpass = real_getpass
        loop.reset()  # reset the singleton


def start_server(port: Optional[int] = None) -> None:
    """Start a simple test web server.

    Run a Flask development server on HOST:PORT with subprocess.
    All output is captured and thrown away.

    The parent process returns the URL on which the server is running.
    """
    global _url, __popen__

    if port is None:
        port = int(os.environ.get("TWILL_TEST_PORT", PORT))

    if START:
        log = LOG or os.devnull
        out = open(log, "w", buffering=1, encoding="utf-8")  # noqa: SIM115
        print(  # noqa: T201
            "Starting:", sys.executable, "tests/server.py", Path.cwd()
        )
        __popen__ = subprocess.Popen(  # noqa: S603
            [sys.executable, "-u", "server.py"],
            stderr=subprocess.STDOUT,
            stdout=out,
        )
        time.sleep(SLEEP)  # wait until the server is up and running
        print("The server has been started.")  # noqa: T201

    _url = f"http://{HOST}:{port}/"


def stop_server() -> None:
    """Stop a previously started test web server."""
    global _url, __popen__

    if __popen__:
        if START:
            try:
                __popen__.terminate()
            except Exception as error:  # noqa: BLE001
                print("ERROR:", error)  # noqa: T201
                print("Could not stop the server.")  # noqa: T201
            else:
                print("The server has been stopped.")  # noqa: T201
        _url = __popen__ = None



"""Extension functions for parsing sys.argv.

Commands:

   get_args -- load all command-line arguments after the last --
      into $arg1...$argN.
"""

from twill import log, namespaces, shell

__all__ = ["get_args"]


def get_args(require: int = 0) -> None:
    """>> get_args [<require>]

    Load the command line arguments after the last '--' into $arg1...$argN,
    optionally requiring at least 'require' such arguments.
    """
    global_dict, _local_dict = namespaces.get_twill_glocals()

    require = int(require)

    if len(shell.twill_args) < require:
        from twill.errors import TwillAssertionError  # noqa: PLC0415

        given = len(shell.twill_args)
        raise TwillAssertionError(
            f"too few arguments; {given} rather than {require}"
        )

    if shell.twill_args:
        for n, arg in enumerate(shell.twill_args, 1):
            global_dict[f"arg{n}"] = arg
        n = len(shell.twill_args)
        log.info("get_args: loaded %d args as $arg1..$arg%d.", n, n)
    else:
        log.info("no arguments to parse!")



"""Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options.
For a full list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Path setup --------------------------------------------------------------

import sys
from pathlib import Path

sys.path.append(str(Path(__file__).resolve().parent.parent / "src"))

# -- Project information -----------------------------------------------------


def project_version() -> str:
    """Fetch version from pyproject.toml file."""
    # this also works when the package is not installed
    with open("../pyproject.toml", encoding="utf-8") as toml_file:
        for line in toml_file:
            if line.startswith("version ="):
                return line.split("=")[1].strip().strip('"')
    raise RuntimeError("Cannot determine project version")


project = "twill"
author = "C. Titus Brown, Ben R. Taylor, Christoph Zwerschke et al"
copyright = "2025, " + author

# The full version, including alpha/beta/rc tags
version = release = project_version()

language = "en"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = []

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages. See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']



import pytest

from twill import parse

from .utils import execute_script


def test_variable_substitution():
    fut = parse.variable_substitution
    args = ({"foo": 7}, {"bar": 13, "baz": "wall"})
    assert (
        fut("${foo} * ${bar} bottles on the ${baz}!", *args)
        == "7 * 13 bottles on the wall!"
    )
    assert fut("${foo * bar}", *args) == str(7 * 13)
    with pytest.raises(ZeroDivisionError):
        fut("${1/0}", {}, {})


def test_variables_script(url: str):
    execute_script("test_variables.twill", initial_url=url)



"""Shared test configuration for pytest."""

from io import StringIO
from typing import Generator

import pytest

from . import utils


@pytest.fixture(scope="session")
def url() -> Generator[str, None, None]:
    """Start a server and returns its URL."""
    utils.cd_test_dir()
    utils.start_server()

    url = utils.get_url()

    from twill import set_output  # noqa: PLC0415
    from twill.commands import find, go  # noqa: PLC0415

    set_output()
    try:
        go(url)
        find("These are the twill tests")
    except Exception as error:
        raise RuntimeError(
            """
***
Hello! The twill test server is not running or cannot be reached;
please free port 8080 (or set TWILL_TEST_PORT to something else),
and clear your proxy settings too!
***
"""
        ) from error

    yield url

    utils.stop_server()
    utils.pop_test_dir()


@pytest.fixture
def output() -> Generator[StringIO, None, None]:
    """Get output from the test."""
    from twill import set_output  # noqa: PLC0415

    with StringIO() as output:
        set_output(output)
        yield output
    set_output()



import pytest

from twill import commands
from twill.errors import TwillException


def test(url: str):
    commands.reset_browser()
    commands.go(url)

    commands.go("/two_forms")
    commands.find(" NO FORM ")

    with pytest.raises(TwillException):
        commands.submit()
    with pytest.raises(TwillException):
        commands.submit("1")

    commands.fv("1", "item", "foo")
    commands.submit()
    commands.find(" FORM=1 ITEM=foo ")

    commands.fv("2", "item", "bar")
    commands.submit()
    commands.find(" FORM=2 ITEM=bar ")

    with pytest.raises(TwillException):
        commands.submit()

    commands.submit("1", "1")
    commands.find(" FORM=1 ITEM= ")

    commands.submit("1", "2")
    commands.find(" FORM=2 ITEM= ")

    with pytest.raises(TwillException):
        commands.submit("1", "3")

    commands.fv("1", "item", "foo")
    commands.fv("2", "item", "bar")
    commands.submit()
    commands.find(" FORM=2 ITEM=bar ")

    commands.fv("2", "item", "bar")
    commands.fv("1", "item", "foo")
    commands.submit()
    commands.find(" FORM=1 ITEM=foo ")



import pytest

from twill import commands
from twill.errors import TwillException

from .utils import execute_script


def test(url: str):
    commands.show()
    commands.show("html")
    commands.show("links")
    with pytest.raises(TwillException, match='Cannot show "nonsense".'):
        commands.show("nonsense")

    execute_script("test_show.twill", initial_url=url)



"""Same as test_basic, but using the command interpreter."""

import sys
from pathlib import Path

import pytest

from twill import __url__ as twill_url
from twill import __version__ as twill_version
from twill.errors import TwillNameError

from .utils import execute_shell, test_dir

python_version = sys.version.split(None, 1)[0]


def test_shell_specific(url: str):
    execute_shell("test_shell.twill", initial_url=url, fail_on_unknown=True)
    text = Path(test_dir, "test_shell.out").read_text()
    lines = [line for line in text.splitlines() if line and "===" not in line]
    expected_lines = [
        f"twill version: {twill_version}",
        f"Python Version: {python_version}",
        f"See {twill_url} for more info.",
        "What do YOU think the command 'help' does?!?",
        "Help for command exit:",
        ">> exit [<code>]",
        'Exit twill with given exit code (default 0, "no error").',
        "Print version information.",
        "Imported extension module 'shell_test'.",
        "testing extension",
        "raising errors",
        "ERROR: The flag has not been set",
        "ERROR: The flag has been set",
    ]

    assert lines == expected_lines


def test_shell_fail(url: str):
    with pytest.raises(TwillNameError):
        execute_shell(
            "test_shell_fail.twill",
            initial_url=url,
            fail_on_unknown=True,
        )


def test_most_commands(url: str):
    inp = "unique1\nunique2\n"
    execute_shell("test_basic.twill", inp, initial_url=url)


def teardown_module():
    for filename in "test_basic.cookies", "test_basic.out", "test_shell.out":
        Path(test_dir, filename).unlink(missing_ok=True)



import sys
from io import StringIO
from pathlib import Path

import pytest

import twill.parse
from twill import commands, namespaces
from twill.errors import TwillAssertionError, TwillNameError

from .utils import execute_script, test_dir

_log_commands = None


def setup_module():
    global _log_commands
    _log_commands = twill.parse.log_commands(True)  # noqa: FBT003


def teardown_module():
    twill.parse.log_commands(_log_commands)


def test(url: str):
    # capture output
    fp = StringIO()
    twill.set_output(fp)

    twill.parse.execute_string("code 200", initial_url=url)

    # from file
    execute_script("test_go.twill", initial_url=url)

    twill.set_output(None)
    assert fp.getvalue()

    path = Path(test_dir, "test_go.twill")
    with open(path) as inp:
        stdin, sys.stdin = sys.stdin, inp
        try:
            execute_script("-", initial_url=url)  # from stdin
        finally:
            sys.stdin = stdin

    # from parse.execute_file
    twill.parse.execute_file("test_go_exit.twill", initial_url=url)

    # also test some failures

    stderr, sys.stderr = sys.stderr, StringIO()
    try:
        twill.set_err_out(sys.stderr)
        # failed assert in a script
        with pytest.raises(TwillAssertionError):
            twill.parse.execute_file("test_go_fail.twill", initial_url=url)

        commands.go(url)
        with pytest.raises(TwillAssertionError):
            commands.code(400)

        # no such command (NameError)
        with pytest.raises(TwillNameError):
            twill.parse.execute_file("test_go_fail2.twill", initial_url=url)
    finally:
        sys.stderr = stderr

    namespaces.new_local_dict()
    gd, ld = namespaces.get_twill_glocals()

    commands.go(url)

    with pytest.raises(TwillAssertionError):
        twill.parse.execute_command("url", ("not this",), gd, ld, "anony")

    with pytest.raises(TwillAssertionError):
        commands.follow("no such link")

    with pytest.raises(TwillAssertionError):
        commands.find("no such link")

    with pytest.raises(TwillAssertionError):
        commands.notfind("Hello")

    with pytest.raises(SystemExit):
        twill.parse.execute_command("exit", ("0",), gd, ld, "anony")



"""Implementation of the TwillBrowser."""

import pickle
import re
from contextlib import suppress
from http import HTTPStatus
from typing import (
    IO,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
)
from urllib.parse import urljoin

from httpx import (
    BasicAuth,
    Client,
    ConnectError,
    Cookies,
    Headers,
    InvalidURL,
    Timeout,
    WSGITransport,
)

from . import __version__, log
from .errors import TwillException
from .utils import (
    CheckboxGroup,
    FieldElement,
    FormElement,
    HtmlElement,
    InputElement,
    Link,
    RadioGroup,
    Response,
    ResultWrapper,
    UrlWithRealm,
    get_equiv_refresh_interval,
    html_to_tree,
    print_form,
    trunc,
    unique_match,
)

__all__ = ["browser"]


def _set_http_connection_debuglevel(level: int) -> None:
    """Set the debug level for the connection pool."""
    from http.client import HTTPConnection  # noqa: PLC0415

    HTTPConnection.debuglevel = level


class TwillBrowser:
    """A simple, stateful browser."""

    user_agent = f"TwillBrowser/{__version__}"

    def __init__(
        self,
        base_url: str = "",
        app: Optional[Callable[..., Any]] = None,
        follow_redirects: bool = True,  # noqa: FBT001, FBT002
        verify: Union[bool, str] = False,  # noqa: FBT001, FBT002
        timeout: Union[None, float, Timeout] = 10,
    ) -> None:
        """Initialize the twill browser.

        Optionally, you can send requests to a WSGI app instead over the
        network, and you can specify a base URL for all requests.
        The "follow_redirects" parameter has the default value True so that
        the browser by default automatically follows all redirects.
        The "verify" argument can be used to specify whether or how server
        certificates shall be verified; this can also be a CA bundle path.
        In the "timeout" argument you can specify the timeout in seconds.
        """
        self.reset(
            app=app,
            base_url=base_url,
            follow_redirects=follow_redirects,
            verify=verify,
            timeout=timeout,
        )

    def _assert_result_for(self, what: str) -> ResultWrapper:
        if not self.result:
            raise TwillException(f"Cannot get {what} since there is no page.")
        return self.result

    @property
    def debug_level(self) -> int:
        return self._debug_level

    @debug_level.setter
    def debug_level(self, level: int) -> None:
        _set_http_connection_debuglevel(level)
        self._debug_level = level

    def close(self) -> None:
        try:
            client = self._client
        except AttributeError:
            pass
        else:
            client.close()
            del self.result
            del self.last_submit_button
            del self.first_error
            del self._client
            del self._form
            del self._form_files
            del self._auth
            del self._post_load_hooks
            del self._history

    def reset(
        self,
        base_url: str = "",
        app: Optional[Callable[..., Any]] = None,
        follow_redirects: bool = True,  # noqa: FBT001, FBT002
        verify: Union[bool, str] = False,  # noqa: FBT001,FBT002
        timeout: Union[None, float, Timeout] = 10,
    ) -> None:
        """Reset the browser.

        Optionally, you can send requests to a WSGI app instead over the
        network, and you can specify a base URL for all requests.
        The "follow_redirects" parameter has the default value True so that
        the browser by default automatically follows all redirects.
        The "verify" argument can be used to specify whether or how server
        certificates shall be verified; this can also be a CA bundle path.
        In the "timeout" argument you can specify the timeout in seconds.
        """
        self.close()

        self.result: Optional[ResultWrapper] = None
        self.last_submit_button: Optional[InputElement] = None
        self.first_error: Optional[str] = None

        # whether meta refresh will be displayed
        self.show_refresh = False

        # debug level to be used for the connection pool
        self._debug_level = 0

        # Client stores cookies
        self._client = Client(
            base_url=base_url,
            follow_redirects=follow_redirects,
            verify=verify,
            timeout=timeout,
            transport=WSGITransport(app=app) if app else None,
        )

        # A lxml FormElement, None until a form is selected
        # replaces self._browser.form from mechanize
        self._form: Optional[FormElement] = None
        self._form_files: Dict[str, IO] = {}

        # A dict of BasicAuth from httpx, keyed off URL
        self._auth: Dict[UrlWithRealm, BasicAuth] = {}

        # callables to be called after each page load.
        self._post_load_hooks: List[Callable] = []

        self._history: List[ResultWrapper] = []

        # set default headers
        self.reset_headers()

    @property
    def creds(self) -> Dict[UrlWithRealm, BasicAuth]:
        """Get the credentials for basic authentication."""
        return self._auth

    def add_creds(self, url: UrlWithRealm, user: str, password: str) -> None:
        """Set the credentials for basic authentication."""
        self._auth[url] = BasicAuth(user, password)

    def go(self, url: str) -> None:
        """Visit given URL."""
        try_urls: List[str] = []
        if "://" in url:
            try_urls.append(url)
        else:  # URL does not have a schema
            # if this is a relative URL, then assume that we want to tack it
            # onto the end of the current URL
            current_url = self.url
            if current_url:
                try_urls.append(urljoin(current_url, url))
            # if this is an absolute URL, it may be just missing the 'http://'
            # at the beginning, try fixing that (mimic browser behavior)
            if not url.startswith((".", "/", "?")):
                try_urls.append(f"http://{url}")
                try_urls.append(f"https://{url}")
        for try_url in try_urls:
            try:
                self._journey("open", try_url)
            except (
                OSError,
                ConnectError,
                InvalidURL,
                UnicodeError,
            ) as error:
                log.info("cannot go to '%s': %s", try_url, error)
            else:
                break
        else:
            raise TwillException(f"cannot go to '{url}'")
        log.info("==> at %s", self.url)

    def reload(self) -> None:
        """Tell the browser to reload the current page."""
        self._journey("reload")
        log.info("==> reloaded")

    def back(self) -> None:
        """Return to previous page, if possible."""
        try:
            self._journey("back")
            log.info("==> back to %s", self.url)
        except TwillException:
            log.warning("==> back at empty page")

    @property
    def code(self) -> int:
        """Get the HTTP status code received for the current page."""
        return self._assert_result_for("status code").http_code

    @property
    def encoding(self) -> Optional[str]:
        """Get the encoding used by the server for the current page."""
        return None if self.result is None else self.result.encoding

    @property
    def html(self) -> str:
        """Get the HTML for the current page."""
        return self._assert_result_for("HTML").text

    @property
    def dump(self) -> bytes:
        """Get the binary content of the current page."""
        return self._assert_result_for("content dump").content

    @property
    def title(self) -> Optional[str]:
        return self._assert_result_for("title").title

    @property
    def url(self) -> Optional[str]:
        """Get the URL of the current page."""
        return self.result.url if self.result else None

    def find_link(self, pattern: str) -> Optional[Link]:
        """Find the first link matching the given regular expression pattern.

        The pattern is searched in the URL and in the link text.
        """
        return self._assert_result_for("links").find_link(pattern)

    def find_links(self, pattern: str) -> Optional[List[Link]]:
        """Find all links matching the given regular expression pattern.

        The pattern is searched in the URL and in the link text.
        """
        return self._assert_result_for("links").find_links(pattern)

    def follow_link(self, link: Union[str, Link]) -> None:
        """Follow the given link."""
        self._journey("follow_link", link)
        log.info("==> at %s", self.url)

    @property
    def headers(self) -> Headers:
        """Return the request headers currently used by the browser."""
        return self._client.headers

    def reset_headers(self) -> None:
        """Reset the request headers currently used by the browser."""
        self.headers.clear()
        self.headers.update(
            {"Accept": "text/html; */*", "User-Agent": self.user_agent}
        )

    @property
    def response_headers(self) -> Headers:
        """Get the headers returned with the current page."""
        return self._assert_result_for("headers").headers

    @property
    def agent_string(self) -> Optional[str]:
        """Get the user agent string."""
        agent = self.headers.get("User-Agent")
        if isinstance(agent, bytes):
            agent = agent.decode()
        return agent

    @agent_string.setter
    def agent_string(self, agent: str) -> None:
        """Set the user agent string to the given value."""
        self.headers["User-Agent"] = agent

    @property
    def timeout(self) -> Union[None, float, Timeout]:
        """Get the request timeout in seconds."""
        timeout = self._client.timeout
        if timeout.connect == timeout.read == timeout.write == timeout.pool:
            return timeout.connect or None
        return timeout

    @timeout.setter
    def timeout(self, timeout: Union[None, float, Timeout]) -> None:
        """Set the request timeout in seconds."""
        self._client.timeout = timeout

    def show_forms(self) -> None:
        """Pretty-print all forms on the page.

        Include the global form (form elements outside <form> pairs)
        as forms[0] if present.
        """
        for n, form in enumerate(self.forms, 1):
            print_form(form, n)

    def show_links(self) -> None:
        """Pretty-print all links on the page."""
        info = log.info
        links = self.links
        if links:
            info("\nLinks (%d links total):\n", len(links))
            for n, link in enumerate(links, 1):
                info("\t%d. %s ==> %s", n, trunc(link.text, 40), link.url)
            info("")
        else:
            info("\n** no links **\n")

    def show_history(self) -> None:
        """Pretty-print the history of links visited."""
        info = log.info
        history = self._history
        if history:
            info("\nHistory (%d pages total):\n", len(history))
            for n, page in enumerate(history, 1):
                info("\t%d. %s", n, page.url)
            info("")
        else:
            info("\n** no history **\n")

    @property
    def links(self) -> List[Link]:
        """Return a list of all links on the page."""
        return self._assert_result_for("links").links

    @property
    def history(self) -> List[ResultWrapper]:
        """Return a list of all pages visited by the browser."""
        return self._history

    @property
    def forms(self) -> List[FormElement]:
        """Return a list of forms on the page.

        This includes the global form at index 0 if present.
        """
        return self._assert_result_for("forms").forms

    def form(self, name: Union[str, int] = 1) -> Optional[FormElement]:
        """Return the first form that matches the given form name."""
        return self._assert_result_for("form").form(name)

    def form_field(
        self,
        form: Optional[FormElement] = None,
        name_or_num: Union[str, int] = 1,
    ) -> FieldElement:
        """Return the control that matches the given field name.

        Must be a *unique* regex/exact string match, but the returned
        control can also be a CheckboxGroup or RadioGroup list.

        Raises a TwillException if no such field or multiple fields are found.
        """
        if form is None:
            form = self._form
            if form is None:
                raise TwillException("Must specify a form for the field")
        inputs = form.inputs
        found_multiple = False

        name = name_or_num if isinstance(name_or_num, str) else None

        if name:
            if name in form.fields:
                match_name = [c for c in inputs if c.name == name]
                if len(match_name) > 1:
                    if all(
                        getattr(c, "type", None) == "checkbox"
                        for c in match_name
                    ):
                        return CheckboxGroup(
                            cast("List[InputElement]", match_name)
                        )
                    if all(
                        getattr(c, "type", None) == "radio" for c in match_name
                    ):
                        return RadioGroup(
                            cast("List[InputElement]", match_name)
                        )
            else:
                match_name = None

            # test exact match to id
            match_id = [c for c in inputs if c.get("id") == name]
            if match_id:
                if unique_match(match_id):
                    return match_id[0]
                found_multiple = True

            # test exact match to name
            if match_name:
                if unique_match(match_name):
                    return match_name[0]
                found_multiple = True

        num = name_or_num if isinstance(name_or_num, int) else None
        if num is None and name and name.isdigit():
            with suppress(ValueError):
                num = int(name)

        # test field index
        if num is not None:
            with suppress(IndexError):
                return list(inputs)[num - 1]

        if name:
            # test regex match
            regex = re.compile(name)
            match_name = [c for c in inputs if c.name and regex.search(c.name)]
            if match_name:
                if unique_match(match_name):
                    return match_name[0]
                found_multiple = True

            # test field values
            match_value = [c for c in inputs if c.value == name]
            if match_value:
                if len(match_value) == 1:
                    return match_value[0]
                found_multiple = True

        # error out
        if found_multiple:
            raise TwillException(f'multiple matches to "{name_or_num}"')
        raise TwillException(f'no field matches "{name_or_num}"')

    def add_form_file(self, field_name: str, fp: IO) -> None:
        self._form_files[field_name] = fp

    def clicked(self, form: FormElement, control: FieldElement) -> None:
        """Record a 'click' in a specific form."""
        if self._form != form:
            # construct a function to choose a particular form;
            # select_form can use this to pick out a precise form.
            self._form = form
            self.last_submit_button = None
        # record the last submit button clicked.
        if getattr(control, "type", None) in ("submit", "image"):
            self.last_submit_button = cast("InputElement", control)

    def submit(
        self,
        field_name: Optional[Union[str, int]] = None,
        form_name: Optional[Union[str, int]] = None,
    ) -> None:
        """Submit the last or specified form using the given field."""
        forms = self.forms
        if not forms:
            raise TwillException("There are no forms on this page.")

        ctl: Optional[InputElement] = None

        form = self._form if form_name is None else self.form(form_name)
        if form is None:
            if len(forms) > 1:
                raise TwillException(
                    "There is more than one form on this page;"
                    " therefore you must specify a form explicitly"
                    " or select one (use 'fv') before submitting."
                )
            form = forms[0]

        action = form.action or ""
        if "://" not in action:
            form.action = urljoin(self.url, action)

        # no field name?  see if we can use the last submit button clicked...
        if field_name is None:
            if form is not self._form or self.last_submit_button is None:
                # get first submit button in form.
                submits = [
                    c
                    for c in form.inputs
                    if getattr(c, "type", None) in ("submit", "image")
                ]
                if submits:
                    ctl = cast("InputElement", submits[0])
            else:
                ctl = self.last_submit_button
        else:
            # field name given; find it
            ctl = cast("InputElement", self.form_field(form, field_name))

        # now set up the submission by building the request object that
        # will be sent in the form submission.
        if ctl is None:
            log.debug("Note: submit without using a submit button")
        else:
            log.info(
                "Note: submit is using submit button: name='%s', value='%s'",
                ctl.get("name"),
                ctl.value,
            )

        # Add referer information.  This may require upgrading the
        # request object to have an 'add_unredirected_header' function.
        # @BRT: For now, the referrer is always the current page
        # @CTB: this seems like an issue for further work.
        # Note: We do not set Content-Type from form.attrib.get('enctype'),
        # since httpx does a much better job at setting the proper one.
        headers = {"Referer": self.url}

        payload = form.form_values()
        if ctl is not None:
            name = ctl.get("name")
            if name:
                payload.append((name, ctl.value or ""))
        payload_dict = self._make_payload_dict(payload)

        # now actually GO
        if form.method == "POST":
            if self._form_files:
                log.debug("Submitting files: %r", self._form_files)
                result = self._client.post(
                    form.action,
                    data=payload_dict,
                    headers=headers,
                    files=self._form_files,
                )
            else:
                result = self._client.post(
                    form.action, data=payload_dict, headers=headers
                )
        else:
            result = self._client.get(
                form.action, params=payload_dict, headers=headers
            )

        self._form = None
        self._form_files.clear()
        self.last_submit_button = None
        if self.result is not None:
            self._history.append(self.result)
        self.result = ResultWrapper(result)

    def cookies(self) -> Cookies:
        """Get all cookies from the current client session."""
        return self._client.cookies

    def save_cookies(self, filename: str) -> None:
        """Save cookies into the given file."""
        saved_cookies = [
            (cookie.name, cookie.value, cookie.domain, cookie.path)
            for cookie in self._client.cookies.jar
        ]
        with open(filename, "wb") as f:
            pickle.dump(saved_cookies, f)

    def load_cookies(self, filename: str) -> None:
        """Load cookies from the given file."""
        cookies = Cookies()
        with open(filename, "rb") as f:
            loaded_cookies = pickle.load(f)  # noqa: S301
        for name, value, domain, path in loaded_cookies:
            cookies.set(name, value, domain=domain, path=path)
        self._client.cookies = cookies

    def clear_cookies(self) -> None:
        """Delete all the cookies."""
        self._client.cookies.clear()

    def show_cookies(self) -> None:
        """Pretty-print all the cookies."""
        info = log.info
        cookies = self._client.cookies
        n = len(cookies)
        if n:
            log.info("\nThere are %d cookie(s) in the cookie jar.\n", n)
            for n, cookie in enumerate(cookies, 1):
                info("\t%d. %s", n, cookie)
            info("")
        else:
            log.info("\nThere are no cookies in the cookie jar.\n")

    def decode(self, value: Union[bytes, str]) -> str:
        """Decode a value using the current encoding."""
        if isinstance(value, bytes):
            value = value.decode(self.encoding or "utf-8")
        return value

    def xpath(self, path: str) -> List[HtmlElement]:
        """Evaluate an xpath expression."""
        return self._assert_result_for("xpath").xpath(path)

    def _make_payload_dict(
        self,
        payload: Sequence[Tuple[str, Union[str, bytes]]],
    ) -> Dict[str, Union[str, List[Union[str]]]]:
        """Prepare a payload by decoding bytes and converting to a dict."""
        encoding = self.encoding or "utf-8"
        data: Dict[str, Union[str, List[Union[str]]]] = {}
        for key, value in payload:
            new_value = (
                value if isinstance(value, str) else value.decode(encoding)
            )
            try:
                existing_value = data[key]
            except KeyError:
                data[key] = new_value
            else:
                if isinstance(existing_value, list):
                    existing_value.append(new_value)
                else:
                    data[key] = [existing_value, new_value]
        return data

    @staticmethod
    def _get_meta_refresh(
        response: Response,
    ) -> Tuple[Optional[int], Optional[str]]:
        """Get meta refresh interval and url from a response."""
        try:
            tree = html_to_tree(response.text)
        except ValueError:
            # may happen when there is an XML encoding declaration
            tree = html_to_tree(response.content)
        try:
            content = tree.xpath(  # "refresh" is case-insensitive
                "//meta[translate(@http-equiv,'REFSH','refsh')="
                "'refresh'][1]/@content"
            )[0]
            interval, url = content.split(";", 1)
            interval = int(interval)
            if interval < 0:
                raise ValueError
            url = url.strip().strip('"').strip().strip("'").strip()
            url = url.split("=", 1)
            if url[0].strip().lower() != "url":
                raise IndexError
            url = url[1].strip().strip('"').strip().strip("'").strip()
        except (IndexError, ValueError):
            interval = url = None
        else:
            if "://" not in url:  # relative URL, adapt
                url = str(response.url.join(url))
        return interval, url

    _re_basic_auth = re.compile('Basic realm="(.*)"', re.I)

    def _journey(self, func_name: str, *args, **_kwargs) -> None:
        """Execute the function with the given name and arguments.

        The name should be one of 'open', 'reload', 'back', or 'follow_link'.
        This method then runs that function with the given arguments and turns
        the results into a nice friendly standard ResultWrapper object, which
        is stored as self.result.

        (Idea stolen from Python Browsing Probe (PBP).)
        """
        self._form = None
        self._form_files.clear()
        self.last_submit_button = None

        if func_name == "open":
            url = args[0]

        elif func_name == "follow_link":
            url = args[0]
            with suppress(AttributeError):  # may be already a url
                url = url.url
            if "://" not in url and self.url:
                url = urljoin(self.url, url)

        elif func_name == "reload":
            url = self.url

        elif func_name == "back":
            try:
                self.result = self._history.pop()
            except IndexError as error:
                raise TwillException("Cannot go further back") from error
            return
        else:
            raise TwillException(f"Unknown function {func_name!r}")

        result = self._client.get(url)
        if result.status_code == HTTPStatus.UNAUTHORIZED:
            header = result.headers.get("WWW-Authenticate")
            match_realm = self._re_basic_auth.match(header)
            if match_realm:
                realm = match_realm.group(1)
                auth = self._auth.get((url, realm)) or self._auth.get(url)
                if auth:
                    result = self._client.get(url, auth=auth)

        # handle redirection via meta refresh (not handled in requests)
        refresh_interval = get_equiv_refresh_interval()
        if refresh_interval:
            visited = set()  # break circular refresh chains
            while True:
                interval, url = self._get_meta_refresh(result)
                if not url:
                    break
                if interval and interval >= refresh_interval:
                    (log.info if self.show_refresh else log.debug)(
                        "Meta refresh interval too long: %d", interval
                    )
                    break
                if url in visited:
                    log.warning("Circular meta refresh detected!")
                    break
                (log.info if self.show_refresh else log.debug)(
                    "Meta refresh to new URL: %s", url
                )
                result = self._client.get(url)
                visited.add(url)

        if (
            func_name in ("follow_link", "open")
            # if we're really reloading and just didn't say so, don't store
            and self.result is not None
            and self.result.url != result.url
        ):
            self._history.append(self.result)

        self.result = ResultWrapper(result)


browser = TwillBrowser()  # the global twill browser instance



#!/usr/bin/env python3

"""Flask based test app for twill."""

import os
from base64 import decodebytes
from time import sleep
from typing import Any, MutableMapping, NoReturn, Optional

from flask import (
    Flask,
    Response,
    abort,
    make_response,
    redirect,
    request,
    session,
)

HOST = "127.0.0.1"
PORT = 8080
DEBUG = False
RELOAD = False

app = Flask(__name__)

app.secret_key = "not-a-secret-since-this-is-a-test-app-only"  # noqa: S105


class SessionManager:
    """Session manager."""

    _next_session_id = 1

    @classmethod
    def message(cls, session: MutableMapping) -> str:
        """Create a message with session information."""
        sid = cls.session_id(session) or "undefined"
        visit = session.get("visit", 0)
        user = session.get("user", "guest")
        return f"""\
<html>
<head>
    <title>Hello, world!</title>
</head>
<body>
    <p>Hello, world!</p>
    <p>These are the twill tests.</p>
    <p>Your session ID is {sid}; this is visit #{visit}.</p>
    <p>You are logged in as {user}.</p>
    <p>
        <a href="increment">increment</a> |
        <a href="incrementfail">incrementfail</a>
    </p>
    <p><a href="logout">log out</a></p>
    <p>
        (<a href="test spaces">test spaces</a> /
        <a href="test_spaces">test spaces2</a>)
    </p>
</body>
</html>
"""

    @classmethod
    def session_id(cls, session: MutableMapping) -> Optional[int]:
        """Get the session ID or set it if it does not exist."""
        sid = session.get("sid")
        if sid is None:
            session["sid"] = cls._next_session_id
            cls._next_session_id += 1
        return sid


message = SessionManager.message


# HTML helpers


def field(type_: str = "text", name: str = "", value: str = "") -> str:
    """Create an HTML field."""
    return f'<input type="{type_}" name="{name}" value="{value}">'


def par(text: str = "") -> str:
    """Create an HTML paragraph."""
    return f"<p>{text}</p>"


# Flask routes


@app.route("/")
def view_root() -> str:
    """Show index page."""
    return message(session)


@app.route("/exception")
def view_exception() -> NoReturn:
    """Raise a server error."""
    raise RuntimeError("500 error -- fail out!")


@app.route("/test_spaces")
@app.route("/test spaces")
def view_test_spaces() -> str:
    """Test spaces."""
    return "success"


@app.route("/sleep")
def view_sleep() -> str:
    """Test timeouts."""
    sleep(0.5)
    return "sorry for the delay"


@app.route("/increment")
def view_increment() -> str:
    """Visit session."""
    session["visit"] = session.get("visit", 0) + 1
    return message(session)


@app.route("/incrementfail")
def view_incrementfail() -> NoReturn:
    """Visit session with failure."""
    session["visit"] = session.get("visit", 0) + 1
    raise RuntimeError(message(session))


@app.route("/simpleform", methods=["GET", "POST"])
def view_simpleform() -> str:
    """Test non-existing submit button."""
    s1, s2 = field(name="n"), field(name="n2")
    values = par(" ".join(request.form.values()))
    return f'{values}</p><form method="POST">{s1} {s2}</form>'


@app.route("/getform")
def view_getform() -> str:
    """Test form with get method."""
    s = field("hidden", name="n", value="v")
    return f'<form method="GET">{s}<input type="submit"></form>'


@app.route("/login", methods=["GET", "POST"])
def view_login() -> Any:
    """Test login."""
    if request.method == "POST":
        username = request.form.get("username")
        if username:
            session["user"] = username
            return redirect("./")

    login = field(name="username", value="")
    s = field("submit", name="submit", value="submit me")
    s2 = field("submit", name="nosubmit2", value="don't submit")
    img = '<input type=image name="submit you" src="DNE.gif">'

    return (
        '<form method="POST"><p>Log in:</p>'
        f"<p>{login}</p><p>{s2} {s}</p><p>{img}</p></form>"
    )


@app.route("/logout")
def view_logout() -> Any:
    """Test logout."""
    session.clear()
    return redirect("./")


@app.route("/upload_file", methods=["GET", "POST"])
def view_upload_file() -> str:
    """Test file upload."""
    if request.method == "POST" and "upload" in request.files:
        return request.files["upload"].read()
    return """\
<form enctype="multipart/form-data" method="POST">
<input type="file" name="upload">
<input type="submit" value="submit">
</form>"""


@app.route("/formpostredirect", methods=["GET", "POST"])
def view_formpostredirect() -> Any:
    """Test redirect after a form POST."""
    if request.method != "POST":
        return """\
<form method="POST" enctype="multipart/form-data">
<input type="text" name="test">
<input type="submit" value="submit" name="submit">
</form>
"""
    return redirect("./")


@app.route("/display_post", methods=["POST"])
def view_display_post() -> str:
    """Show the form items."""
    return "".join(par(f"{k!s}: {v!r}") for k, v in request.form.items())


@app.route("/display_environ")
def view_display_environ() -> str:
    """Show the environment variables."""
    return "".join(par(f"{k!s}: {v!r}") for k, v in request.environ.items())


@app.route("/echo")
def view_echo() -> str:
    """Show query parameters."""
    show = ", ".join(f"{k}={v}" for k, v in request.args.items()) or "nothing"
    return f"<html><body>{show}</body></html>"


@app.route("/plaintext")
def view_plaintext() -> Response:
    """Test plain text response."""
    response = make_response("hello, world")
    response.headers["Content-Type"] = "text/plain"
    return response


@app.route("/xml")
def view_xml() -> Response:
    """Test XML response."""
    response = make_response(
        '<?xml version="1.0" encoding="utf-8" ?><foo>bår</foo>'
    )
    response.headers["Content-Type"] = "text/xml"
    return response


@app.route("/restricted")
def view_restricted() -> str:
    """Test restricted access."""
    if "user" not in session:
        response = make_response("you must have a username", 403)
        abort(response)
    return "you made it!"


@app.route("/http_auth")
def view_http_auth() -> str:
    """Test restricted access using HTTP authentication."""
    login = passwd = None
    ha = request.environ.get("HTTP_AUTHORIZATION")
    if ha:
        auth_type, auth_string = ha.split(None, 1)
        if auth_type.lower() == "basic":
            auth_string = decodebytes(auth_string.encode("utf-8"))
            login_bytes, passwd_bytes = auth_string.split(b":", 1)
            login = login_bytes.decode("utf-8")
            passwd = passwd_bytes.decode("utf-8")
            if (login, passwd) != ("test", "password"):
                passwd = None

    if passwd:
        print(f"Successful login as {login}")
    elif login:
        print(f"Invalid login attempt as {login}")
    else:
        print("Access has been denied")
    print()
    if not passwd:
        response = make_response(
            "you are not authorized to access this resource", 401
        )
        response.headers["WWW-Authenticate"] = 'Basic realm="Protected"'
        abort(response)
    return "you made it!"


@app.route("/broken_linktext")
def view_broken_linktext() -> str:
    """Get broken link text."""
    return """
<a href="/">
<span>some text</span>
</a>
"""


@app.route("/broken_form_1")
def view_broken_form_1() -> str:
    """Get broken form 1."""
    return """\
<form>
<input type="text" name="blah" value="thus">
"""


@app.route("/broken_form_2")
def view_broken_form_2() -> str:
    """Get broken form 2."""
    return """\
<form>
<table>
<tr><td>
<input name="broken">
</td>
</form>
</tr>
</form>
"""


@app.route("/broken_form_3")
def view_broken_form_3() -> str:
    """Get broken form 3."""
    return """\
<table>
<tr><td>
<input name="broken">
</td>
</form>
</tr>
</form>
"""


@app.route("/broken_form_4")
def view_broken_form_4() -> str:
    """Get broken form 4."""
    return """\
<font>
<INPUT>

<FORM>
<input type="blah">
</form>
"""


@app.route("/broken_form_5")
def view_broken_form_5() -> str:
    """Get broken form 5."""
    return """\
<div id="loginform">
<form method="post" name="loginform" action="ChkLogin">
<h3>ARINC Direct Login</h3>
<br/>
<strong>User ID</strong><br/>
<input name="username" id="username" type="text" style="width:80%"><br/>
<strong>Password</strong><br/>
<input name="password" type="password" style="width:80%"><br/>
<div id="buttonbar">
<input value="Login" name="login" class="button" type="submit">
</div>
</form>
</div>
"""


@app.route("/test_checkbox", methods=["GET", "POST"])
def view_test_checkbox() -> str:
    """Test single checkbox."""
    s = ""
    if request.method == "POST" and "checkboxtest" in request.form:
        value = request.form["checkboxtest"]

        s += par(f"CHECKBOXTEST: =={value}==")

    return f"""\
{s}
<form method="POST">

<input type="checkbox" name="checkboxtest" value="True">
<input type="hidden" name="checkboxtest" value="False">

<input type="submit" value="post">
</form>
"""


@app.route("/test_checkboxes", methods=["GET", "POST"])
def view_test_checkboxes() -> str:
    """Test multiple checkboxes."""
    s = ""
    if request.method == "POST" and "checkboxtest" in request.form:
        value = ",".join(request.form.getlist("checkboxtest"))

        s += par(f"CHECKBOXTEST: =={value}==")

    return f"""\
{s}
<form method="POST">
<input type="checkbox" name="checkboxtest" value="one">
<input type="checkbox" name="checkboxtest" value="two">
<input type="checkbox" name="checkboxtest" value="three">
<input type="submit" value="post">
</form>
"""


@app.route("/test_simple_checkbox", methods=["GET", "POST"])
def view_test_simple_checkbox() -> str:
    """Test simple checkbox."""
    s = ""
    if request.method == "POST" and "checkboxtest" in request.form:
        value = request.form["checkboxtest"]

        s += par(f"CHECKBOXTEST: =={value}==")

    return f"""\
{s}
<form method="POST">

<input type="checkbox" name="checkboxtest">

<input type="submit" value="post">
</form>
"""


@app.route("/test_radiobuttons", methods=["GET", "POST"])
def view_test_radiobuttons() -> str:
    """Test radio buttons."""
    s = ""
    if request.method == "POST" and "radiobuttontest" in request.form:
        value = request.form["radiobuttontest"]

        s += par(f"RADIOBUTTONTEST: =={value}==")

    return f"""\
{s}
<form method="POST">
<input type="radio" name="radiobuttontest" value="one">
<input type="radio" name="radiobuttontest" value="two">
<input type="radio" name="radiobuttontest" value="three">
<input type="submit" value="post">
</form>
"""


@app.route("/testformaction", methods=["POST"])
def view_testformaction() -> str:
    """Test form actions."""
    keys = sorted(k for k in request.form if request.form[k])
    return "==" + " AND ".join(keys) + "=="


@app.route("/testform", methods=["GET", "POST"])
def view_testform() -> str:
    """Test form."""
    s = ""
    if not request.form:
        s = "NO FORM"

    if request.form and "selecttest" in request.form:
        values = " AND ".join(request.form.getlist("selecttest"))
        s += par(f"SELECTTEST: =={values}==")

    if request.form:
        items = []
        for name in ("item", "item_a", "item_b", "item_c"):
            if request.form.get(name):
                value = request.form[name]
                items.append(f"{name}={value}")
        values = " AND ".join(items)
        s += par(f"NAMETEST: =={values}==")

    return f"""\
{s}
<form method="POST" id="the_form">
<select name="selecttest" multiple>
<option>val</option>
<option value="selvalue1">value1</option>
<option value="selvalue2">value2</option>
<option value="selvalue3">value3</option>
<option value="test.value3">testme.val</option>
<option value="Test.Value4">testme4.val</option>
</select>

<input type="text" name="item">
<input type="text" name="item_a">
<input type="text" name="item_b">
<input type="text" name="item_c">

<input type="text" id="some_id">

<input type="submit" value="post" id="submit_button">
</form>
"""


@app.route("/multisubmitform", methods=["GET", "POST"])
def view_multisubmitform() -> str:
    """Test form with multiple submit buttons."""
    s1 = field("submit", "sub_a", value="sub_a")
    s2 = field("submit", "sub_b", value="sub_b")

    s = ""
    if request.method == "POST":
        used = False
        if "sub_a" in request.form:
            used = True
            s += "used_sub_a"
        if "sub_b" in request.form:
            used = True
            s += "used_sub_b"

        if not used:
            raise RuntimeError("Not button was used.")

        referer = request.environ.get("HTTP_REFERER")
        if referer:
            s += par(f"referer: {referer}")

    return f'<form method="POST">{s} {s1} {s2}</form>'


@app.route("/two_forms", methods=["GET", "POST"])
def view_two_forms() -> str:
    """Test two forms."""
    if request.form:
        form = request.form.get("form")
        item = request.form.get("item")
        s = f"FORM={form} ITEM={item}"
    else:
        s = "NO FORM"

    return f"""\
<h1>Two Forms</h1>
<p>== {s} ==</p>
<form method="POST" id="form1">
<input type="text" name="item">
<input type="hidden" name="form" value="1">
<input type="submit" value="post">
</form>
<form method="POST" id="form2">
<input type="text" name="item">
<input type="hidden" name="form" value="2">
<input type="submit" value="post">
</form>
"""


@app.route("/test_global_form", methods=["GET", "POST"])
def view_test_global_form() -> str:
    """Test the global form."""
    return """
<html>
<head>
<title>Broken</title>
</head>
<body>
<div>
<input name="global_form_entry" type="text">
<input name="global_entry_2" type="text">
</div>

<form name="login" method="post" action="display_post">
<input type="text" name="hello">
<input type="hidden" name="form" value="1">
<input type="submit">
</form>

<form name="login" method="post" action="display_post">
<input type="text" name="hello">
<input type="hidden" name="form" value="2">
<input type="submit">
</form>

</body>
</html>
"""


@app.route("/test_refresh")
def view_test_refresh() -> str:
    """Test simple refresh."""
    return """\
<meta http-equiv="refresh" content="2; url=./login">
hello, world.
"""


@app.route("/test_refresh2")
def view_test_refresh2() -> str:
    """Test refresh with upper case."""
    return """\
<META HTTP-EQUIV="REFRESH" CONTENT="2; URL=./login">
hello, world.
"""


@app.route("/test_refresh3")
def view_test_refresh3() -> str:
    """Test circular refresh."""
    return """\
<meta http-equiv="refresh" content="2; url=./test_refresh3">
hello, world.
"""


@app.route("/test_refresh4")
def view_test_refresh4() -> str:
    """Test refresh together with similar meta tags."""
    return """\
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
<head>
<title>o2.ie</title>
<meta http-equiv="refresh" content="0;URL=/login">
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">
</head>
<body>
</body>
</html>
hello, world.
"""


@app.route("/test_refresh5")
def view_test_refresh5() -> str:
    """Check for situation where given URL is quoted."""
    return """\
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<html>
<head>
<title>o2.ie</title>
<meta http-equiv="refresh" content="0;'URL=/login'">
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">
</head>
<body>
</body>
</html>
hello, world.
"""


if __name__ == "__main__":
    port = int(os.environ.get("TWILL_TEST_PORT", PORT))
    print(f"starting twill test server on port {port}.")
    try:
        app.run(host=HOST, port=port, debug=DEBUG, use_reloader=RELOAD)
    except KeyboardInterrupt:
        print("Keyboard interrupt ignored.")



from pathlib import Path

import pytest

from twill import utils
from twill.errors import TwillException


def test_is_hidden_filename():
    is_hidden_filename = utils.is_hidden_filename
    assert not is_hidden_filename("foo")
    assert is_hidden_filename(".foo")
    assert not is_hidden_filename(".foo/bar")
    assert is_hidden_filename("foo/.bar")


def test_is_twill_filename():
    is_twill_filename = utils.is_twill_filename
    assert not is_twill_filename("foo")
    assert is_twill_filename("foo.twill")
    assert not is_twill_filename(".foo.twill")


def test_make_boolean():
    make_boolean = utils.make_boolean
    assert make_boolean(True)  # noqa: FBT003
    assert not make_boolean(False)  # noqa: FBT003
    assert make_boolean("true")
    assert not make_boolean("false")
    assert make_boolean(1)
    assert not make_boolean(0)
    assert make_boolean("1")
    assert not make_boolean("0")
    assert make_boolean("+")
    assert not make_boolean("-")
    with pytest.raises(TwillException):
        make_boolean("no")


def test_make_twill_filename():
    make_twill_filename = utils.make_twill_filename
    assert make_twill_filename("test_foo") == "test_foo"
    assert make_twill_filename("../tests/test_foo") == str(
        Path("../tests/test_foo")
    )
    assert make_twill_filename("test_basic") == "test_basic.twill"
    assert make_twill_filename("../tests/test_basic") == str(
        Path("../tests/test_basic.twill")
    )


def test_trunc():
    trunc = utils.trunc
    assert trunc("hello, world!", 12) == "hello, w ..."
    assert trunc("hello, world!", 13) == "hello, world!"



"""A simple set of extensions to manage post-load requirements for pages.

Commands:

   require       -- turn on post-load requirements; either 'success' or
                    'links_ok'.

   no_require    -- turn off requirements.

   skip_require  -- for the next page visit, skip requirements processing.

   flush_visited -- flush the list of already visited pages
                    (for links checking)
"""

from twill import browser, commands, log

__all__ = ["flush_visited", "no_require", "require", "skip_require"]

_requirements = []  # what requirements to satisfy


class Ignore:
    once: bool = False  # reset after each hook call
    always: bool = False  # never reset


def skip_require() -> None:
    """>> skip_require

    Skip the post-page-load requirements.
    """
    Ignore.once = True


def require(what: str) -> None:
    """>> require <what>

    After each page is loaded, require that 'what' be satisfied.  'what'
    can be:
      * 'success' -- HTTP return code is 200
      * 'links_ok' -- all the links on the page load OK (see 'check_links'
                      extension module)
    """
    # install the post-load hook function.
    hooks = browser._post_load_hooks  # noqa: SLF001
    if _require_post_load_hook not in hooks:
        log.debug("INSTALLING POST-LOAD HOOK")
        hooks.append(_require_post_load_hook)

    # add the requirement.
    if what not in _requirements:
        log.debug("Adding requirement: %s", what)
        _requirements.append(what)


def no_require() -> None:
    """>> no_require

    Remove all post-load requirements.
    """
    hooks = browser._post_load_hooks  # noqa: SLF001
    hooks = [fn for fn in hooks if fn != _require_post_load_hook]
    browser._post_load_hooks = hooks  # noqa: SLF001

    _requirements.clear()


def flush_visited() -> None:
    """>> flush_visited

    Flush the list of pages successfully visited already.
    """
    from .check_links import good_urls  # noqa: PLC0415

    good_urls.clear()


def _require_post_load_hook(action: str, *_args, **_kwargs) -> None:
    """Post load hook function to be called after each page is loaded.

    See TwillBrowser._journey() for more information.
    """
    if action == "back":  # do nothing on a 'back'
        return

    if Ignore.once or Ignore.always:
        Ignore.once = False
        return

    for what in _requirements:
        if what == "success":
            log.debug("REQUIRING success")
            commands.code(200)

        elif what == "links_ok":
            from .check_links import check_links, good_urls  # noqa: PLC0415

            Ignore.always = True
            log.debug("REQUIRING functioning links")
            log.debug("(already visited:)")
            log.debug("\n\t".join(sorted(good_urls)))

            try:
                check_links()
            finally:
                Ignore.always = False



"""Test a boatload of miscellaneous functionality."""

import sys
from io import StringIO

import pytest

from twill import browser, commands
from twill.errors import TwillException


def test():
    assert browser is not None
    for attr in ("go", "reset", "submit"):
        assert hasattr(browser, attr)

    # reset
    commands.reset_browser()
    assert browser is not None
    for attr in ("go", "reset", "submit"):
        assert hasattr(browser, attr)

    with pytest.raises(TwillException):  # no page and thus no status code yet
        assert browser.code

    with pytest.raises(TwillException):  # no page and thus no form yet
        browser.submit()

    stderr, sys.stderr = sys.stderr, StringIO()
    try:
        with pytest.raises(TwillException):
            browser.go("http://0.0.0.0")  # URL parses, but is invalid
    finally:
        sys.stderr = stderr

    with pytest.raises(SystemExit):
        commands.exit()

    commands.reset_browser()
    commands.showhistory()

    with pytest.raises(TwillException):  # no page, cannot tidy yet
        commands.tidy_ok()

    with pytest.raises(TwillException):  # no page, cannot show yet
        commands.show()

    commands.debug("http", "1")
    commands.debug("http", "0")
    commands.debug("http", "+")
    commands.debug("http", "-")

    commands.debug("commands", "0")
    commands.debug("commands", "1")
    with pytest.raises(TwillException):
        commands.debug("nada", "1")

    commands.config()

    commands.config("readonly_controls_writeable")

    commands.config("readonly_controls_writeable", 1)
    commands.config("readonly_controls_writeable", "on")
    with pytest.raises(TwillException):
        commands.config("readonly_controls_writeable", "nada")

    commands.run("print('Hello!')")



from twill import browser, commands


def test_links_parsing(url: str):  # noqa: ARG001
    """Test parsing a link text inside a span."""
    commands.go("/broken_linktext")
    # make sure link text is found even if it is nested
    commands.follow("some text")


def test_fixing_forms(url: str):
    """Test parsing of broken HTML forms."""
    commands.go(url)

    commands.go("/broken_form_1")
    assert len(browser.forms) == 1, "can fix form 1"

    commands.go("/broken_form_2")
    assert len(browser.forms) == 1, "can fix form 2"

    commands.go("/broken_form_3")
    assert len(browser.forms) == 1, "can fix form 3"

    commands.go("/broken_form_4")
    assert len(browser.forms) == 2, "can fix form 4"

    commands.go("/broken_form_5")
    assert len(browser.forms) == 1, "can fix form 5"

    assert set(browser.form().inputs.keys()) == {
        "username",
        "password",
        "login",
    }, "should get proper fields"



"""Test the WSGI support."""

from typing import Callable, Dict, Generator, List, Sequence, Tuple

from twill import browser, commands

"""Intercept HTTP connections that use
`requests <http://docs.python-requests.org/en/latest/>`_.
"""

app_was_hit = set()

Environ = Dict[str, str]
WriteCallable = Callable[[bytes], None]
Headers = List[Tuple[str, str]]
StartResponse = Callable[[str, Headers], WriteCallable]


def simple_app(
    _environ: Environ,
    start_response: StartResponse,
) -> Sequence[bytes]:
    """Simplest possible application object."""
    status = "200 OK"
    response_headers = [("Content-type", "text/plain")]
    start_response(status, response_headers)
    app_was_hit.add("simple_app")
    return (b"Hello, World!",)


def write_app(
    _environ: Environ,
    start_response: StartResponse,
) -> Sequence[bytes]:
    """Simple application using a legacy write callable.

    See https://peps.python.org/pep-3333/#the-write-callable.
    """
    status = "200 OK"
    response_headers = [("Content-type", "text/plain")]
    write = start_response(status, response_headers)
    app_was_hit.add("write_app")
    write(b"Hello, ")
    return (b"World!",)


class IteratorApp:
    """Simple application using a custom iterator."""

    content = (b"Hello, world!",)

    def __init__(
        self,
        environ: Environ,
        start_response: StartResponse,
    ) -> None:
        """Create this application."""
        self.environ = environ
        self.start_response = start_response

    def __iter__(self) -> Generator[bytes, None, None]:
        status = "200 OK"
        response_headers = [("Content-type", "text/plain")]
        self.start_response(status, response_headers)
        app_was_hit.add(self.__class__.__name__)
        yield from self.content


def test_simple_app():
    app_was_hit.clear()
    browser.reset(app=simple_app)
    try:
        assert not app_was_hit
        commands.go("http://localhost:8080/")
        commands.show()
        commands.find("Hello, World!")
        assert "simple_app" in app_was_hit
    finally:
        browser.reset()


def test_write_app():
    app_was_hit.clear()
    browser.reset(app=write_app)
    try:
        assert not app_was_hit
        commands.go("http://localhost:8080/")
        commands.show()
        # next line may be added again once this has been merged:
        # https://github.com/encode/httpx/pull/2920
        # commands.find("Hello, World!")  # noqa: ERA001
        assert "write_app" in app_was_hit
    finally:
        browser.reset()


def test_iterator_app():
    browser.reset(app=IteratorApp)
    try:
        commands.go("http://localhost:80/")
        commands.show()
        commands.find("Hello, world!")
        commands.notfind("!Hello")
        assert "IteratorApp" in app_was_hit
    finally:
        browser.reset()



"""Test the utils.run_tidy function.

This doesn't test to see if tidy was actually run; all it does is make sure
that the function runs without error...
"""

import pytest

from twill import utils
from twill.commands import config
from twill.errors import TwillException

bad_html = """<a href="test">you</a> <b>hello."""


def setup_module():
    config("require_tidy", 1)


def teardown_module():
    config("require_tidy", 0)


def test_bad_html():
    output, errors = utils.run_tidy(bad_html)
    assert errors
    output, errors = utils.run_tidy(output)
    assert not errors


def test_no_tidylib():
    tidylib, utils.tidylib = utils.tidylib, None
    try:
        with pytest.raises(TwillException, match="PyTidyLib is not installed"):
            utils.run_tidy(bad_html)
    finally:
        utils.tidylib = tidylib


def test_no_tidylib_but_not_required():
    config("require_tidy", 0)
    tidylib, utils.tidylib = utils.tidylib, None
    try:
        output, errors = utils.run_tidy(bad_html)
    finally:
        utils.tidylib = tidylib

    assert output is None
    assert errors is None
    config("require_tidy", 1)


def test_tidy_options():
    good_content = "<h1>Hello, World!</h1>"
    output, errors = utils.run_tidy(good_content)
    assert errors
    config("tidy_show_body_only", 1)
    output, errors = utils.run_tidy(good_content)
    assert not errors
    config("tidy_show_body_only", 0)
    output, errors = utils.run_tidy(good_content)
    assert errors


if __name__ == "__main__":
    setup_module()
    test_bad_html()
    teardown_module()



"""A command-line interpreter for twill.

This is an implementation of a command-line interpreter based on the
'Cmd' class in the 'cmd' package of the default Python distribution.
"""

import os
import sys
import traceback
from argparse import ArgumentParser
from cmd import Cmd
from contextlib import suppress
from pathlib import Path
from textwrap import dedent
from typing import IO, Any, Callable, List, Optional

from . import (
    __url__,
    __version__,
    commands,
    execute_file,
    log,
    log_levels,
    namespaces,
    parse,
    set_log_level,
    set_output,
    shutdown,
)
from .browser import browser
from .utils import Singleton, gather_filenames

readline: Any
try:
    import readline
except ImportError:
    # may not exist on some operating systems
    readline = None

__all__ = ["main"]

python_version = sys.version.split(None, 1)[0]

version_info = f"""
twill version: {__version__}
Python Version: {python_version}

See {__url__} for more info.
"""


def make_cmd_fn(cmd: str) -> Callable[[str], None]:
    """Make a command function.

    Dynamically define a twill shell command function based on an imported
    function name.  (This is where the twill.commands functions actually
    get executed.)
    """

    def do_cmd(rest_of_line: str, cmd: str = cmd) -> None:
        global_dict, local_dict = namespaces.get_twill_glocals()

        args = []
        if rest_of_line.strip():
            try:
                args = parse.arguments.parseString(rest_of_line)[0]
                args = parse.process_args(args, global_dict, local_dict)
            except Exception as error:  # noqa: BLE001
                log.error("\nINPUT ERROR: %s\n", error)
                return

        try:
            parse.execute_command(
                cmd, args, global_dict, local_dict, "<shell>"
            )
        except SystemExit:
            raise
        except Exception as error:  # noqa: BLE001
            log.error("\nERROR: %s\n", error)
            log.debug(traceback.format_exc())

    return do_cmd


def make_help_cmd(cmd: str, docstring: str) -> Callable[[str], None]:
    """Make a help command function.

    Dynamically define a twill shell help function for the given
    command/docstring.
    """

    def help_cmd(message: str = docstring, cmd: str = cmd) -> None:
        message = message.strip()
        try:
            title, details = message.split("\n", 1)
        except ValueError:
            pass
        else:  # dedent the details (for Python < 3.13)
            details = dedent(details)
            message = "\n".join((title, details))
        max_width = max(
            7 + len(cmd),
            *(len(line.rstrip()) for line in message.splitlines()),
        )
        indent = "=" * max_width
        info = log.info
        info("\n%s", indent)
        info("\nHelp for command %s:\n", cmd)
        info(message)
        info("\n%s\n", indent)

    return help_cmd


def add_command(cmd: str, docstring: str) -> None:
    """Add a command with given docstring to the shell."""
    shell = get_command_shell()
    if shell:
        shell.add_command(cmd, docstring)


class TwillCommandLoop(Singleton, Cmd):
    """The command-line interpreter for twill commands.

    This is a Singleton object: you can't create more than one
    of shell at a time.

    Note: most of the do_ and help_ functions are dynamically created
    by the metaclass.
    """

    def __init__(
        self,
        stdin: Optional[IO[str]] = None,
        initial_url: Optional[str] = None,
        *,
        fail_on_unknown: bool = False,
    ) -> None:
        Cmd.__init__(self, stdin=stdin)

        self.use_rawinput = stdin is None

        # initialize a new local namespace.
        namespaces.new_local_dict()

        # import readline history, if available/possible.
        if readline is not None:
            with suppress(OSError):
                readline.read_history_file(".twill-history")

        # fail on unknown commands? for test-shell, primarily.
        self.fail_on_unknown = fail_on_unknown

        # handle initial URL argument
        if initial_url:
            commands.go(initial_url)

        self._set_prompt()

        self.names: List[str] = []

        global_dict, _local_dict = namespaces.get_twill_glocals()

        # add all of the commands from twill
        for command in parse.command_list:
            fn = global_dict.get(command)
            self.add_command(command, fn.__doc__)

    def add_command(self, command: str, docstring: str) -> None:
        """Add the given command into the lexicon of all commands."""
        do_name = f"do_{command}"
        do_cmd = make_cmd_fn(command)
        setattr(self, do_name, do_cmd)

        if docstring:
            help_cmd = make_help_cmd(command, docstring)
            help_name = f"help_{command}"
            setattr(self, help_name, help_cmd)

        self.names.append(do_name)

    def get_names(self) -> List[str]:
        """Return the list of commands."""
        return self.names

    def complete_form_value(
        self, text: str, line: str, _begin: int, _end: int
    ) -> List[str]:
        """Command arg completion for the form_value command.

        The twill command has the following syntax:

        form_value <form_name> <field_name> <value>
        """
        _cmd, args = parse.parse_command(line + ".", {}, {})
        if args:
            place = len(args) if args else 0
            if place == 1:
                return self.provide_form_name(text)
            if place == 2:  # noqa: PLR2004
                form_name = args[0]
                return self.provide_field_name(form_name, text)
        return []

    complete_fv = complete_form_value  # alias

    @staticmethod
    def provide_form_name(prefix: str) -> List[str]:
        """Provide the list of form names on the given page."""
        names = []
        forms = browser.forms
        for form in forms:
            form_id = form.attrib.get("id")
            if form_id and form_id.startswith(prefix):
                names.append(form_id)
                continue
            name = form.attrib.get("name")
            if name and name.startswith(prefix):
                names.append(name)
        return names

    @staticmethod
    def provide_field_name(form_name: str, prefix: str) -> List[str]:
        """Provide the list of fields for the given form_name or number."""
        names = []
        form = browser.form(form_name)
        if form is not None:
            for field in form.inputs:
                field_id = field.attrib.get("id")
                if field_id and field_id.startswith(prefix):
                    names.append(field_id)
                    continue
                name = field.name
                if name and name.startswith(prefix):
                    names.append(name)
        return names

    def _set_prompt(self) -> None:
        """Set the prompt to the current page."""
        url = browser.url
        if url is None:
            url = " *empty page* "
        self.prompt = f"current page: {url}\n>> "

    def precmd(self, line: str) -> str:
        """Run before each command; save."""
        return line

    def postcmd(self, stop: bool, line: str) -> bool:  # noqa: ARG002, FBT001
        """Run after each command; set prompt."""
        self._set_prompt()
        return stop

    def default(self, line: str) -> None:
        """Run when an unknown command is executed."""
        # empty lines ==> emptyline(); here we just want to remove
        # leading whitespace.
        line = line.strip()

        # look for command
        global_dict, local_dict = namespaces.get_twill_glocals()
        cmd, args = parse.parse_command(line, global_dict, local_dict)

        # ignore comments & empty stuff
        if cmd is None:
            return

        try:
            parse.execute_command(
                cmd, args or (), global_dict, local_dict, "<shell>"
            )
        except SystemExit:
            raise
        except Exception as error:
            log.error("\nERROR: %s\n", error)
            if self.fail_on_unknown:
                raise

    def emptyline(self) -> Any:
        """Handle empty lines (by ignoring them)."""

    @staticmethod
    def do_EOF(*_args: str) -> None:  # noqa: N802
        """Exit on CTRL-D."""
        if readline is not None:
            with suppress(OSError):
                readline.write_history_file(".twill-history")
        raise SystemExit

    @staticmethod
    def help_help() -> None:
        """Show help for the help command."""
        log.info("\nWhat do YOU think the command 'help' does?!?\n")

    @staticmethod
    def do_version(*_args: str) -> None:
        """Show the version number of twill."""
        log.info(version_info)

    @staticmethod
    def help_version() -> None:
        """Show help for the version command."""
        log.info("\nPrint version information.\n")

    def do_exit(self, *_args: str) -> None:
        """Exit the twill shell."""
        raise SystemExit

    @staticmethod
    def help_exit() -> None:
        """Show help for the exit command."""
        log.info("\nExit twill.\n")

    do_quit = do_exit
    help_quit = help_exit


def get_command_shell() -> Optional[TwillCommandLoop]:
    """Get the command shell."""
    return getattr(TwillCommandLoop, "__it__", None)


twill_args: List[str] = []  # contains sys.argv *after* last '--'
interactive = False  # 'True' if interacting with user


def main(  # noqa: C901, PLR0912, PLR0915
    argv: Optional[List[str]] = None,
) -> None:
    """Run as shell script."""
    global interactive

    # show the shorthand name for usage
    if argv is None:
        argv = sys.argv
    if argv[0].endswith("-script.py"):
        argv[0] = argv[0].rsplit("-", 1)[0]

    # make sure that the current working directory is in the path
    if "" not in sys.path:
        sys.path.append("")

    parser = ArgumentParser()
    add = parser.add_argument

    add(
        "-d",
        "--dump-html",
        action="store",
        dest="dumpfile",
        help="dump HTML to this file on error",
    )
    add(
        "-f",
        "--fail",
        action="store_true",
        dest="fail",
        help="fail exit on first file to fail",
    )
    add(
        "-i",
        "--interactive",
        action="store_true",
        dest="interactive",
        help="drop into an interactive shell (after running files)",
    )
    add(
        "-l",
        "--loglevel",
        action="store",
        dest="loglevel",
        help="set the logging level",
    )
    add(
        "-n",
        "--never-fail",
        action="store_true",
        dest="never_fail",
        help="continue executing scripts past errors",
    )
    add(
        "-o",
        "--output",
        action="store",
        dest="outfile",
        help="print log to output file",
    )
    add(
        "-q",
        "--quiet",
        action="store_true",
        dest="quiet",
        help="do not show normal output",
    )
    add(
        "-u",
        "--url",
        action="store",
        dest="url",
        help="start at the given URL before each script",
    )
    add(
        "-v",
        "--version",
        action="store_true",
        dest="show_version",
        help="show version information and exit",
    )
    add(
        "-w",
        "--show-error-in-browser",
        action="store_true",
        dest="show_browser",
        help="show dumped HTML in a web browser ",
    )
    add(
        "scripts",
        metavar="SCRIPT",
        nargs="*",
        help="the twill script to execute",
    )

    # parse arguments
    raw_args = argv[1:]
    if "--" in raw_args:
        for last in reversed(range(len(raw_args))):
            if raw_args[last] == "--":
                twill_args[:] = raw_args[last + 1 :]
                raw_args = raw_args[:last]
                break

    args = parser.parse_args(raw_args)

    if args.show_version:
        log.info(version_info)
        sys.exit(0)

    quiet = args.quiet
    show_browser = args.show_browser
    dump_file = args.dumpfile
    out_file = args.outfile
    log_level = args.loglevel
    scripts = args.scripts
    interactive = args.interactive or not scripts

    if out_file:
        out_file = out_file.lstrip("=").lstrip() or None
        if out_file == "-":
            out_file = None

    if interactive and (quiet or out_file or dump_file or show_browser):
        sys.exit("Interactive mode is incompatible with -q, -o, -d and -w")

    if show_browser and (not dump_file or dump_file == "-"):
        sys.exit("Please also specify a dump file with -d")

    if log_level:
        log_level = log_level.lstrip("=").lstrip() or None
        if log_level and log_level.upper() not in log_levels:
            log_level_names = ", ".join(sorted(log_levels))
            sys.exit(f"Valid log levels are: {log_level_names}")
        set_log_level(log_level)

    if quiet:
        output = open(os.devnull, "w")  # noqa: SIM115
    elif out_file:
        try:
            output = open(out_file, "w")  # noqa: SIM115
        except OSError as error:
            sys.exit(f"Invalid output file '{out_file}': {error}")
    else:
        output = None

    set_output(output)

    # first find and run any scripts put on the command line

    failed = False
    if scripts:
        success = []
        failure = []

        filenames = gather_filenames(scripts)
        dump = None

        for filename in filenames:
            try:
                interactive = False
                execute_file(
                    filename,
                    initial_url=args.url,
                    never_fail=args.never_fail,
                )
                success.append(filename)
            except Exception as error:
                if dump_file:
                    dump = browser.dump
                if args.fail:
                    raise
                if browser.first_error:
                    log.error("\nFirst error: %s", browser.first_error)
                log.error("\n*** ERROR: %s", error)
                log.debug(traceback.format_exc())
                failure.append(filename)

        log.info("--")
        if dump:
            if dump_file == "-":
                log.info(
                    "HTML when error was encountered:\n\n%s\n--", dump.strip()
                )
            else:
                try:
                    with open(dump_file, "wb") as f:
                        f.write(dump)
                except OSError as e:
                    log.error("Could not dump to %s: %s\n", dump_file, e)
                else:
                    log.info("HTML has been dumped to %s\n", dump_file)

        log.info(
            "%d of %d files SUCCEEDED.",
            len(success),
            len(success) + len(failure),
        )
        if failure:
            log.error("Failed:\n\t%s", "\n\t".join(failure))
            failed = True

        if dump and show_browser:
            import webbrowser  # noqa: PLC0415

            url = Path(dump_file).absolute().as_uri()
            log.debug("Running web browser on %s", url)
            webbrowser.open(url)

    # if no scripts to run or -i is set, drop into an interactive shell

    if interactive:
        welcome_msg = "" if scripts else "\n -= Welcome to twill =-\n"

        shell = TwillCommandLoop(initial_url=args.url)

        while True:
            try:
                shell.cmdloop(welcome_msg)
            except KeyboardInterrupt:
                sys.stdout.write("\n")
                break

            welcome_msg = ""

    shutdown()

    if failed:
        sys.exit(1)
    sys.exit(0)


if __name__ == "__main__":
    main()



import os
from pathlib import Path

from twill.utils import gather_filenames


def test_gather_dir():
    test_dir = Path(__file__).parent / "test_gather"
    cwd = Path.cwd()

    os.chdir(test_dir)
    try:
        files = gather_filenames((os.curdir,))
        if os.sep != "/":
            files = [f.replace(os.sep, "/") for f in files]
        assert sorted(files) == sorted(
            (
                "./00-testme/x-script.twill",
                "./01-test/b.twill",
                "./02-test2/c.twill",
                "./02-test2/02-subtest/d.twill",
            )
        ), files
    finally:
        os.chdir(cwd)



"""An extension function to iterate over a list of comma-separated values.

Function 'csv_iterate' reads a file containing one or more rows of
comma-separated columns, assigns them to col1...colN, and, for each row,
executes the given twill script.
"""

import csv

from twill import execute_file, log, namespaces

__all__ = ["csv_iterate"]


def csv_iterate(file_name: str, script_name: str) -> None:
    """>> csv_iterate <csv_file> <script>

    For each line in <csv_file>, read in a list of comma-separated values,
    put them in $col1...$colN, and execute <script>.
    """
    global_dict, local_dict = namespaces.get_twill_glocals()

    with open(file_name, encoding="utf-8") as csv_file:
        reader = csv.reader(csv_file)
        for i, row in enumerate(reader, 1):
            log.debug("csv_iterate: on row %d of %s", i, file_name)
            for j, col in enumerate(row, 1):
                global_dict[f"col{j}"] = col

    execute_file(script_name, no_reset=True)



import twill
from twill import commands
from twill.utils import Link


def test(url: str):
    commands.reset_browser()
    browser = twill.browser
    commands.go(url)

    link = browser.find_link("logout")
    assert isinstance(link, Link)
    assert link == ("log out", "logout")

    link = browser.find_link("log out")
    assert isinstance(link, Link)
    assert link == ("log out", "logout")

    link = browser.find_link("log ?out")
    assert isinstance(link, Link)
    assert link == ("log out", "logout")

    assert browser.find_link("log off") is None

    assert browser.find_link("Logout") is None

    link = browser.find_link("test.*")
    assert isinstance(link, Link)
    assert link == ("test spaces", "test spaces")

    links = browser.find_links("logout")
    assert all(isinstance(link, Link) for link in links)
    assert links == [("log out", "logout")]

    links = browser.find_links("log out")
    assert all(isinstance(link, Link) for link in links)
    assert links == [("log out", "logout")]

    links = browser.find_links("log? out")
    assert all(isinstance(link, Link) for link in links)
    assert links == [("log out", "logout")]

    assert browser.find_links("log off") == []

    assert browser.find_links("Logout") == []

    links = browser.find_links("test.*")
    assert all(isinstance(link, Link) for link in links)
    assert links == [
        ("test spaces", "test spaces"),
        ("test spaces2", "test_spaces"),
    ]

    links = browser.find_links(".*")
    assert all(isinstance(link, Link) for link in links)
    assert len(links) == 5



"""Test the unit-test support framework using (naturally) a unit test."""

import os
import time
from pathlib import Path

import twill.unit

from .server import app
from .utils import test_dir

HOST = "127.0.0.1"
PORT = 8080
SLEEP = 0.25
DEBUG = False
RELOAD = False


def run_server() -> None:
    """Function to run the server"""
    port = int(os.environ.get("TWILL_TEST_PORT", PORT))
    app.run(host=HOST, port=port, debug=DEBUG, use_reloader=RELOAD)


def just_wait() -> None:
    """Just wait instead of starting the server."""
    while True:
        time.sleep(SLEEP or 1)


def test(url: str) -> None:
    """Test wrapper for unit support."""
    # abspath to the script
    script = str(Path(test_dir, "test_unit_support.twill"))

    # the port that shall be used for running the server
    port = int(os.environ.get("TWILL_TEST_PORT", PORT))

    # create a TestInfo object
    test_info = twill.unit.TestInfo(script, run_server, HOST, port, SLEEP)

    if test_info.url == url:
        # server already running, run a dummy server function that just waits
        test_info = twill.unit.TestInfo(script, just_wait, HOST, port, SLEEP)

    # run the tests!
    twill.unit.run_test(test_info)



from io import StringIO

from .utils import execute_script


def test(url: str, output: StringIO):
    execute_script("test_global_form.twill", initial_url=url)
    out = output.getvalue()
    assert "Form #1" in out
    assert "Form name=login (#2)" in out
    assert "Form name=login (#3)" in out



from time import sleep

import pytest
from httpx import ReadTimeout

from twill import commands


def test(url: str):
    commands.reset_browser()
    commands.go(url)
    commands.go("/sleep")
    commands.find("sorry for the delay")
    commands.timeout(0.25)  # do not wait for server
    with pytest.raises(ReadTimeout):
        commands.go("/sleep")
    sleep(0.25)  # now wait for server to catch up
    commands.timeout(900)
    commands.timeout(10)
    commands.timeout(0)
    commands.timeout()
    commands.reset_browser(url)
    commands.go(url)



"""
Run through the test-basic script.

This should really be broken out into multiple sub scripts...
"""

from pathlib import Path

from .utils import execute_script, test_dir


def test(url: str):
    inp = "unique1\nunique2\n"

    execute_script("test_basic.twill", inp, initial_url=url)


def teardown_module():
    for filename in "test_basic.cookies", "test_basic.out":
        Path(test_dir, filename).unlink(missing_ok=True)



import pytest

from twill import browser, commands, namespaces
from twill.errors import TwillException


def test_switch_buttons(url: str):
    namespaces.new_local_dict()
    commands.reset_browser()

    with pytest.raises(TwillException):
        browser.title  # noqa: B018

    commands.go(url)
    commands.go("/test_radiobuttons")

    commands.submit()
    assert "RADIOBUTTONTEST" not in browser.html

    for x in ("one", "two", "three"):
        commands.fv("1", "radiobuttontest", x)
        commands.submit()
        assert f"RADIOBUTTONTEST: =={x}==" in browser.html



"""Used in test_shell, to test default command execution and extensions."""

from twill.errors import TwillAssertionError

__all__ = ["assert_flag_set", "assert_flag_unset", "set_flag", "unset_flag"]


class Flag:
    """Global flag value."""

    value: bool = False


def set_flag() -> None:
    """Set the flag."""
    Flag.value = True


def unset_flag() -> None:
    """Unset the flag."""
    Flag.value = False


def assert_flag_set() -> None:
    """Assert that the flag has been set."""
    if not Flag.value:
        raise TwillAssertionError("The flag has not been set")


def assert_flag_unset() -> None:
    """Assert that the flag has not been set."""
    if Flag.value:
        raise TwillAssertionError("The flag has been set")



import pytest

import twill
from twill import commands, namespaces
from twill.errors import TwillAssertionError, TwillException

from .utils import execute_script


def test(url: str):
    # test empty page title
    namespaces.new_local_dict()
    commands.reset_browser()
    browser = twill.browser
    with pytest.raises(TwillException):
        browser.title  # noqa: B018

    # now test a few special cases
    commands.go(url)
    assert browser.title == "Hello, world!"

    commands.go("/login")
    assert browser.title is None

    # test no matching forms
    with pytest.raises(TwillAssertionError):
        commands.fv("2", "submit", "1")

    # test regex match
    commands.fv("1", ".*you", "1")

    # test ambiguous match to value
    commands.go("/testform")
    commands.fv("1", "selecttest", "val")
    commands.fv("1", "selecttest", "value1")
    commands.fv("1", "selecttest", "selvalue1")
    commands.formclear("1")
    commands.showforms()
    with pytest.raises(TwillException):
        commands.fv("1", "selecttest", "value")
    # test ambiguous match to name
    commands.go("/testform")
    with pytest.raises(TwillException):
        commands.fv("1", "item_", "value")

    with pytest.raises(TwillException):
        commands.formfile("1", "selecttest", "null")

    # test the twill script.
    execute_script("test_form.twill", initial_url=url)



"""A twill extension example."""

from twill import log


def test(*args):
    """Test passed arguments."""
    log.info("function test passed %d args", len(args))
    log.info("args are: %s", args)


log.info("---")
log.info("imported function 'test';")
log.info("try 'test hello world!'")
log.info("---")



"""Simple mock implementation of dnspython to test the twill DNS extension."""

import enum
import sys
from typing import Any, Dict, List, Optional


class RdataType(enum.IntEnum):
    """DNS Rdata Type."""

    A = 1
    NS = 2
    CNAME = 5
    MX = 15


mock_records: Dict[RdataType, Dict[str, str]] = {
    RdataType.A: {
        "twill-test-1.ignore.idyll.org": "192.168.1.1",
        "twill-test-2.ignore.idyll.org": "192.168.1.2",
        "twill-test-3.ignore.idyll.org": "192.168.1.1",
    },
    RdataType.CNAME: {
        "twill-test-3.ignore.idyll.org": "twill-test-1.ignore.idyll.org.",
    },
    RdataType.MX: {
        "twill-test-4.ignore.idyll.org": "twill-test-2.ignore.idyll.org.",
    },
    RdataType.NS: {
        "idyll.org": "nsa.idyll.org.",
    },
}


def activate() -> None:
    """Activate the mock dns module."""
    package = sys.modules[__name__]
    sys.modules["dns"] = package
    for module in ["ipv4", "name", "rdatatype", "resolver"]:
        sys.modules[f"dns.{module}"] = package
        setattr(package, module, package)


def inet_aton(text: str) -> str:
    """Convert IPv4 address in text form to network form."""
    try:
        net = "".join(chr(int(x)) for x in text.split("."))
        if len(net) != 4:  # noqa: PLR2004
            raise ValueError
    except (TypeError, ValueError) as error:
        raise OSError(f"invalid ip address {text}") from error
    return net


def from_text(text: str) -> str:
    """Convert text into a Name object."""
    if not text.endswith("."):
        text += "."
    return text


class Result:
    """DNS query result."""

    def __init__(self, qtype: RdataType, value: Any) -> None:
        """Initialize the result."""
        self.value = value
        if qtype == RdataType.A:
            self.address = value
        elif qtype in (RdataType.CNAME, RdataType.NS):
            self.target = value
        elif qtype == RdataType.MX:
            self.exchange = value
        else:
            raise ValueError(f"unknown query type: {qtype}")

    def __str__(self) -> str:
        """Get a string representation of the result."""
        return str(self.value)


Answer = List[Result]


class Resolver:
    """DNS stub resolver."""

    def __init__(self) -> None:
        """Initialize the resolver."""
        self.nameservers: Optional[List] = None

    def resolve(self, qname: str, qtype: RdataType = RdataType.A) -> Answer:
        """Query nameservers to find the answer to the question."""
        if self.nameservers:
            raise ValueError(f"unknown name servers: {self.nameservers}")
        try:
            records = mock_records[qtype]
        except KeyError as error:
            raise ValueError(f"unknown query type: {qtype}") from error
        if qname.endswith("."):
            qname = qname[:-1]
        try:
            results = records[qname]
        except KeyError as error:
            raise ValueError(
                f"unknown query result: {qname} {qtype}"
            ) from error
        if not isinstance(results, list):
            results = [results]  # type: ignore[assignment]
        return [Result(qtype, result) for result in results]



import pytest

import twill
from twill import browser, commands, namespaces
from twill.errors import TwillException


def test_select_multiple(url: str):
    namespaces.new_local_dict()
    commands.reset_browser()

    with pytest.raises(TwillException):
        browser.title  # noqa: B018

    commands.go(url)
    commands.go("/test_checkboxes")

    commands.fv("1", "checkboxtest", "one")
    commands.fv("1", "checkboxtest", "two")
    commands.fv("1", "checkboxtest", "three")

    commands.fv("1", "checkboxtest", "-one")
    commands.fv("1", "checkboxtest", "-two")
    commands.fv("1", "checkboxtest", "-three")

    commands.submit()
    assert "CHECKBOXTEST" not in browser.html

    commands.fv("1", "checkboxtest", "+one")
    commands.fv("1", "checkboxtest", "+two")
    commands.fv("1", "checkboxtest", "+three")

    commands.submit()
    assert "CHECKBOXTEST: ==one,two,three==" in browser.html

    commands.fv("1", "checkboxtest", "-one")
    commands.fv("1", "checkboxtest", "-two")
    commands.fv("1", "checkboxtest", "-three")

    commands.submit()
    assert "CHECKBOXTEST" not in browser.html


def test_select_single(url: str):
    namespaces.new_local_dict()
    commands.reset_browser()
    browser = twill.browser
    with pytest.raises(TwillException):
        browser.title  # noqa: B018

    commands.go(url)
    commands.go("/test_checkboxes")

    # Should not be able to use a bool style for when
    # there are multiple checkboxes
    for x in ("1", "0", "True", "False"):
        with pytest.raises(KeyError):
            commands.fv("1", "checkboxtest", x)



"""Extension functions for easier form filling.

(This module is a dumping ground for features that may ultimately get
added into the main twill command set.)

Commands:

 * fv_match -- fill in *all* fields that match a regex (unlike 'form_value'
        which will complain about multiple matches).  Useful for forms
        with lots of repeated field names -- 'field-1', 'field-2', etc.

 * fv_multi -- fill in multiple form fields at once, e.g.

          fv_multi <form_name> field1=value1 field2=value2 field3=value3

 * fv_multi_sub -- same as 'fv_multi', followed by a 'submit'.
"""

import re

from twill import browser, commands, log, utils

__all__ = ["fv_match", "fv_multi", "fv_multi_match", "fv_multi_sub"]


def fv_match(form_name: str, field_pattern: str, value: str) -> None:
    """>> fv_match <form_name> <field_pattern> <value>

    Set value of *all* form fields with a name that matches the given
    regular expression pattern.

    (Unlike 'form_value' or 'fv', this will not complain about multiple
    matches!)
    """
    form = browser.form(form_name)
    if form is None:
        log.error("no such form '%s'", form_name)
        return

    regex = re.compile(field_pattern)

    matches = [
        ctl for ctl in form.inputs if regex.search(str(ctl.get("name")))
    ]

    if matches:
        log.info("-- matches %d", len(matches))

        n = 0
        for control in matches:
            browser.clicked(form, control)
            if "readonly" in control.attrib:
                continue

            n += 1
            utils.set_form_control_value(control, value)

        log.info("set %d values total", n)


def fv_multi_match(form_name: str, field_pattern: str, *values: str) -> None:
    """>> fv_multi_match <form_name> <field_pattern> <value>...

    Set value of each consecutive form field matching the given pattern with
    the next specified value.  If there are no more values, use the last for
    all remaining form fields.
    """
    form = browser.form(form_name)
    if form is None:
        log.error("no such form '%s'", form_name)
        return

    regex = re.compile(field_pattern)

    matches = [
        ctl for ctl in form.inputs if regex.search(str(ctl.get("name")))
    ]

    if matches:
        log.info("-- matches %d, values %d", len(matches), len(values))

        for n, control in enumerate(matches):
            browser.clicked(form, control)
            if "readonly" in control.attrib:
                continue
            try:
                utils.set_form_control_value(control, values[n])
            except IndexError:
                utils.set_form_control_value(control, values[-1])

        log.info("set %d values total", n)


def fv_multi(form_name: str, *pairs: str) -> None:
    """>> fv_multi <form_name> <pair>...

    Set multiple form fields; each pair should be of the form

        field_name=value

    The pair will be split around the first '=', and
    'fv <form_name> field_name value' will be executed in the order the
    pairs are given.
    """
    for pair in pairs:
        field_name, value = pair.split("=", 1)
        commands.fv(form_name, field_name, value)


def fv_multi_sub(form_name: str, *pairs: str) -> None:
    """>> fv_multi_sub <form_name> <pair>...

    Set multiple form fields (as with 'fv_multi') and then submit().
    """
    for pair in pairs:
        field_name, value = pair.split("=", 1)
        commands.fv(form_name, field_name, value)

    commands.submit()
